import Foundation
import Metal
import MetalKit

// The 256 byte aligned size of our uniform structure
let _alignedUniformsSize = (MemoryLayout<Uniforms>.size & ~0xFF) + 0x100

let _maxBuffersInFlight = 3

class VexilModel : NSObject {
	var _vertices:[[Float]] = []
	private var _vertexBuffer : [MTLBuffer] = []
	private var _vertexSize : [Int] = []

	private var _dynamicUniformBuffer: MTLBuffer
	private var _uniformBufferOffset:Int = 0
	private var _uniformBufferIndex:Int = 0
	private var _uniforms: UnsafeMutablePointer<Uniforms>

	var _visible:Bool = true
	var _translation:simd_float3 = simd_float3(0,0,0)
	var _rotation:simd_float3 = simd_float3(0,0,0)
	var _scaling:simd_float3 = simd_float3(1,1,1)

	private var _colorBuffer: [MTLBuffer] = []
	private var _diffuse: [SIMD4<Float>] = []
	private var _textureName: [String] = []
	private var _texture:[MTLTexture?] = []
	public var _materials: [[Int]] = []
	
	var _matricies:[[Float]] = []
	private var _matrixBuffer: [MTLBuffer] = []
	private var _animationSize: Int = 0
	private var _time: Float = 0
	public var _maxBonesSize: Int = 0
	private var _currentAnimation: Int = 0
	public var _isPlaying: Bool = true
	private var _loop: Bool = true
	public var _animationNames:[String:Int] = [:]

    var _modelViewMatrix: simd_float4x4 = simd_float4x4()
	public var _distance:Float = 0
	private var _distance2:Float = 0
	public var _intersectU:Float = 0
	public var _intersectV:Float = 0

	init?(device:MTLDevice) {
		let uniformBufferSize = _alignedUniformsSize * _maxBuffersInFlight
		guard let buffer = device.makeBuffer(length:uniformBufferSize, options:[MTLResourceOptions.storageModeShared]) else { return nil }
		_dynamicUniformBuffer = buffer
		self._dynamicUniformBuffer.label = "UniformBuffer"
		_uniforms = UnsafeMutableRawPointer(_dynamicUniformBuffer.contents()).bindMemory(to:Uniforms.self, capacity:1)

		super.init()
	}
	
	func initialize(device:MTLDevice) {
		for vert in _vertices {
			let buf = device.makeBuffer(bytes: vert, length: vert.count*MemoryLayout<Float>.size, options: MTLResourceOptions.cpuCacheModeWriteCombined)
			_vertexBuffer.append(buf!)
			_vertexSize.append(vert.count/11)
		}
	}
	
	func setMaterial(texture:String,red:Float,green:Float,blue:Float,alpha:Float) {
		_textureName.append(texture)
		var red2 = red
		if texture.count > 0 {
			red2 = -1
		}
		let diffuse = SIMD4<Float>(red2,green,blue,alpha)
		_diffuse.append(diffuse)
	}
	
	func setColor(device:MTLDevice) {
		for dif in _diffuse {
			let d:[Float] = [dif.x,dif.y,dif.z,dif.w]
			let buf = device.makeBuffer(bytes: d, length: 4*MemoryLayout<Float>.size, options: MTLResourceOptions.cpuCacheModeWriteCombined)
			_colorBuffer.append(buf!)
		}
	}
	
	func setTexture(device:MTLDevice) {
		let textureLoader = MTKTextureLoader(device: device)
		
		let textureLoaderOptions = [
			MTKTextureLoader.Option.textureUsage: NSNumber(value: MTLTextureUsage.shaderRead.rawValue),
			MTKTextureLoader.Option.textureStorageMode: NSNumber(value: MTLStorageMode.`private`.rawValue)
		]
		for name in _textureName {
			if name.count > 0 {
				let path:URL? = Bundle.main.url(forResource: name, withExtension: "")
				var tex:MTLTexture? = nil
				do {
					tex = try textureLoader.newTexture(URL: path!, options: textureLoaderOptions)
				} catch {
				}
				_texture.append(tex)
			} else {
				_texture.append(nil)
			}
		}
	}
	
	func encoder(renderEncoder:MTLRenderCommandEncoder) {
		if !_visible {
			return
		}
		if _isPlaying {
			_time += 30.0/60.0
		}
		if _time >= Float(_animationSize) {
			_time = 0
			if !_loop {
				_isPlaying = false
			}
		}
		
		for i in 0..<_vertexSize.count {
			if _vertexSize[i] <= 0 {
				continue
			}
			renderEncoder.setVertexBuffer(_vertexBuffer[i], offset: 0, index: 0)
			renderEncoder.setVertexBuffer(_matrixBuffer[_currentAnimation], offset: _maxBonesSize * 16 * Int(_time) * MemoryLayout<Float>.size, index: 1)
			renderEncoder.setVertexBuffer(_dynamicUniformBuffer, offset:_uniformBufferOffset, index: 2)
			renderEncoder.setFragmentBuffer(_colorBuffer[_materials[_currentAnimation][i]], offset:0, index: 0)
			if ( _textureName[i].count > 0 ) {
				renderEncoder.setFragmentTexture(_texture[_materials[_currentAnimation][i]], index: 0)
			}
			renderEncoder.drawPrimitives(type: MTLPrimitiveType.triangle, vertexStart: 0, vertexCount: _vertexSize[i])
		}
	}

	func playAnimation(animation: Int, loop: Bool) {
		_isPlaying = true
		_loop = loop
		if animation < 0 {
			_currentAnimation = 0
		}
		else if animation < _matricies.count {
			_currentAnimation = animation
		}
		else
		{
			_currentAnimation = 0
		}
		if _maxBonesSize > 0 {
			_animationSize = _matricies[_currentAnimation].count / (_maxBonesSize*16)
		} else {
			_animationSize = 0
		}
		_time = 0
	}

	func playAnimationByName(animation: String, loop: Bool) {
		_isPlaying = true
		_loop = loop
		let anim = _animationNames[animation]
		if anim != nil {
			_currentAnimation = anim!
		}
		else
		{
			_currentAnimation = 0
		}
		if _maxBonesSize > 0 {
			_animationSize = _matricies[_currentAnimation].count / (_maxBonesSize*16)
		} else {
			_animationSize = 0
		}
		_time = 0
	}

	func stopAnimation() {
		_isPlaying = false
	}
	
	func updateDynamicBufferState() {
		_uniformBufferIndex = (_uniformBufferIndex + 1) % _maxBuffersInFlight
		
		_uniformBufferOffset = _alignedUniformsSize * _uniformBufferIndex
		
		_uniforms = UnsafeMutableRawPointer(_dynamicUniformBuffer.contents() + _uniformBufferOffset).bindMemory(to:Uniforms.self, capacity:1)
	}

	func updateGameState(projectionMatrix:matrix_float4x4,viewMatrix:matrix_float4x4) {
		_uniforms[0].projectionMatrix = projectionMatrix
		
		let translate = Transforms.translate(_translation)
		let rotateX = Transforms.rotateX(degree:_rotation.x)
		let rotateY = Transforms.rotateY(degree:_rotation.y)
		let rotateZ = Transforms.rotateZ(degree:_rotation.z)
		let scale = Transforms.scale(scale: _scaling)
		var matrix = rotateX * scale
		matrix = rotateY * matrix
		matrix = rotateZ * matrix
		matrix = translate * matrix
		matrix = viewMatrix * matrix
		_uniforms[0].modelViewMatrix = matrix
        _modelViewMatrix = matrix
		matrix = matrix.transpose
		let normalMatrix:simd_float3x3 = Transforms.toFloat3x3(f44:matrix.inverse)
		_uniforms[0].normalMatrix = normalMatrix
	}
	
	func setMatrixAnimation(device:MTLDevice) {
		if _maxBonesSize > 0 {
			_animationSize = _matricies[0].count / (_maxBonesSize*16)
		} else {
			_matricies = [[1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1]]
			_animationSize = 0
		}
		for mat in _matricies {
			let buf = device.makeBuffer(bytes: mat, length: mat.count*MemoryLayout<Float>.size, options: MTLResourceOptions.cpuCacheModeWriteCombined)
			_matrixBuffer.append(buf!)
		}
	}
    
	func intersect(origin:simd_float3, dir:simd_float3) -> Bool
	{
		var flg = false
		_distance = 1000000
		for i in 0..<_vertexSize.count {
			for j in stride(from: 0, to: _vertices[i].count, by: 33) {
				let v = _vertices[i]
				let v0:simd_float3 = simd_float3(v[j],v[j+1],v[j+2])
				let v1:simd_float3 = simd_float3(v[j+11],v[j+12],v[j+13])
				let v2:simd_float3 = simd_float3(v[j+22],v[j+23],v[j+24])
				if ( intersectTriangle(origin:origin,dir:dir,v0:v0,v1:v1,v2:v2) )
				{
					if ( _distance2 >= 0 && _distance > _distance2 )
					{
						_distance = _distance2
						flg = true
					}
				}
			}
		}
		return flg
	}
	
	func intersectTriangle(origin:simd_float3, dir:simd_float3, v0:simd_float3, v1:simd_float3, v2:simd_float3)->Bool
	{
		let e1:simd_float3 = v1 - v0
		let e2:simd_float3 = v2 - v0
		let pv:simd_float3 = simd_cross(dir, e2)
		var tv:simd_float3 = simd_float3()
		var qv:simd_float3 = simd_float3()
		let det = simd_dot(e1,pv)
		var t:Float
		var u:Float
		var v:Float
		var inv_det:Float
	
		if (det > (1e-3))
		{
			tv = origin - v0
			u = simd_dot(tv,pv)
			if (u < 0.0 || u > det)
			{
				return false
			}
			qv = simd_cross(tv,e1)
	
			v = simd_dot(dir,qv)
			if (v < 0.0 || u + v > det)
			{
				return false
			}
		}
		else if (det < -(1e-3))
		{
			tv = origin - v0
			u = simd_dot(tv,pv)
			if (u > 0.0 || u < det)
			{
				return false
			}
			qv = simd_cross(tv,e1)
	
			v = simd_dot(dir,qv)
			if (v > 0.0 || u + v < det)
			{
				return false
			}
		}
		else
		{
			return false
		}
	
		inv_det = 1.0 / det
	
		t = simd_dot(e2,qv)
		t *= inv_det
		u *= inv_det
		v *= inv_det
	
		_distance2 = t
		_intersectU = u
		_intersectV = v
	
		return true
	}
    
    func mousePick(x:CGFloat,y:CGFloat,width:CGFloat,height:CGFloat,projection:matrix_float4x4)->Bool
    {
        var vPickRayDir:simd_float3 = simd_float3()
        var vPickRayOrig:simd_float3 = simd_float3()
        var v:simd_float3 = simd_float3()

        v.x =  Float( ( ( 2.0 * x ) / width  ) - 1.0 ) / projection[0][0]
        v.y = -Float( ( ( 2.0 * y ) / height ) - 1.0 ) / projection[1][1]
        v.z = 1.0

        var m:matrix_float4x4
        m = _modelViewMatrix.inverse

        vPickRayDir.x  = v.x*m[0][0] + v.y*m[1][0] + v.z*m[2][0]
        vPickRayDir.y  = v.x*m[0][1] + v.y*m[1][1] + v.z*m[2][1]
        vPickRayDir.z  = v.x*m[0][2] + v.y*m[1][2] + v.z*m[2][2]
        vPickRayOrig.x = m[3][0]
        vPickRayOrig.y = m[3][1]
        vPickRayOrig.z = m[3][2]

        if( intersect(origin:vPickRayOrig,dir:vPickRayDir) )
        {
            return true
        }

        return false
    }
}
